import asyncio
import logging

from aiohttp import ClientConnectorError, ClientSession, WSMsgType

from lona.message_bus.protocol import decode_message, encode_message

logger = logging.getLogger('lona.message_bus.client')


class MessageBusClient:
    def __init__(self, server):
        self.server = server

        self.session = None
        self.websocket = None
        self._running = True

        self.url = self.server.render_string(
            self.server.settings.MESSAGE_CLIENT_URL,
            {
                'server': self.server,
                'settings': self.server.settings,
            },
        )

        self.issuer = self.server.render_string(
            self.server.settings.MESSAGE_CLIENT_ISSUER,
            {
                'server': self.server,
                'settings': self.server.settings,
            },
        )

    def handle_message(self, raw_message):
        logger.debug('raw message received: %s', repr(raw_message))

        message_is_valid, message = decode_message(raw_message)

        if not message_is_valid:
            logger.warning('invalid message skipped: %s', repr(raw_message))

            return

        self.server.middleware_controller.handle_bus_message(
            issuer=message[0],
            topic=message[1],
            params=message[2],
        )

    async def _receive_messages(self):
        self.session = ClientSession()

        while self._running:
            try:
                logger.debug('connecting to %s', self.url)

                self.websocket = await self.session.ws_connect(self.url)

                logger.debug('connected')

                async for message in self.websocket:
                    if message.type == WSMsgType.TEXT:
                        self.server.run_function_async(
                            self.handle_message,
                            message.data,
                        )

                    elif message.type == WSMsgType.PING:
                        await self.websocket.pong()

                    elif message.type in (WSMsgType.CLOSED, WSMsgType.ERROR):
                        await self.websocket.close()

            except asyncio.CancelledError:
                logger.debug('CancelledError')

                return

            except ClientConnectorError:
                logger.error('ConnectionRefusedError: {}'.format(self.url))

            except Exception:
                logger.error(
                    'exception raised',
                    exc_info=True,
                )

            logger.debug('disconnected')

            if self._running:
                logger.debug(
                    'retrying in %ss',
                    self.server.settings.MESSAGE_CLIENT_RETRY_INTERVAL,
                )

                await asyncio.sleep(
                    self.server.settings.MESSAGE_CLIENT_RETRY_INTERVAL,
                )

            else:
                logger.debug('shutting down')

    async def start(self):
        if not self.server.settings.MESSAGE_CLIENT:
            logger.debug('message client disabled')

            return

        logger.debug('start')

        self._running = True
        self.server.loop.create_task(self._receive_messages())

    async def stop(self):
        if not self.server.settings.MESSAGE_CLIENT:
            return

        self._running = False

        if self.session is not None:
            await self.session.close()

        if self.websocket is not None:
            await self.websocket.close()

        logger.debug('stop')

    def send_str(self, string, sync=True):
        try:
            return self.server.run(
                self.websocket.send_str(string),
                sync=sync,
            )

        except ConnectionResetError:
            # this exception gets handled by aiohttp internally and
            # can be ignored

            pass

        except Exception:
            logger.error(
                'exception raised while sending message',
                exc_info=True,
            )

    def send_message(self, topic, issuer=None, params=None):
        issuer = issuer or self.issuer

        message = encode_message(
            issuer=issuer,
            topic=topic,
            params=params,
        )

        self.send_str(message)



import logging
import asyncio

from aiohttp.web import WebSocketResponse, Response
from aiohttp import WSMsgType

from lona.message_bus.protocol import decode_message
from lona.connection import Connection

logger = logging.getLogger('lona.message_bus.broker')


class MessageBusBroker:
    def __init__(self, server):
        self.server = server

        self.connections = []

    async def stop(self, *args, **kwargs):
        logger.debug('stop')

        for connection in self.connections.copy():
            try:
                await connection.close()

            except Exception:
                pass

    async def broadcast(self, raw_message):
        counter = 0

        for connection in self.connections.copy():
            try:
                await connection.send_str(raw_message, sync=False)
                counter += 1

            except ConnectionResetError:
                # this exception gets handled by aiohttp internally and
                # can be ignored

                pass

        logger.debug(
            'message %s broadcasted to %s clients',
            repr(raw_message),
            counter,
        )

    def handle_websocket_message(self, raw_message):
        logger.debug('raw message received: %s', repr(raw_message))

        message_is_valid, message = decode_message(raw_message)

        if not message_is_valid:
            logger.warning('invalid message skipped: %s', repr(raw_message))

            return

        self.server.run_coroutine_sync(
            self.broadcast(raw_message),
            wait=False,
        )

    async def handle_websocket_request(self, http_request):
        # setup websocket
        websocket = WebSocketResponse()
        await websocket.prepare(http_request)

        logger.debug('client %s: connected', hex(id(websocket)))

        connection = Connection(
            server=self.server,
            http_request=http_request,
            websocket=websocket,
        )

        # run middlewares
        middleware_controller = self.server.middleware_controller

        handled, message, middleware = \
            await middleware_controller.handle_message_bus_connection(
                connection=connection,
            )

        # connection got rejected
        if handled:
            logger.debug(
                'client %s: connection rejected by %s',
                hex(id(websocket)),
                repr(middleware),
            )

            if isinstance(message, str):
                try:
                    await connection.send_str(message, sync=False)

                except Exception:
                    pass

            await connection.close()

            return websocket

        # main loop
        self.connections.append(connection)

        try:
            async for message in websocket:
                if message.type == WSMsgType.TEXT:
                    self.server.run_function_async(
                        self.handle_websocket_message,
                        message.data,
                    )

                elif message.type == WSMsgType.PING:
                    await websocket.pong()

                elif message.type in (WSMsgType.CLOSED, WSMsgType.ERROR):
                    break

        except asyncio.CancelledError:
            logger.debug('CancelledError')

        finally:
            logger.debug('client %s disconnected', id(websocket))

            await websocket.close()

            if connection in self.connections:
                self.connections.remove(connection)

        return websocket

    async def handle_http_request(self, http_request):
        if(http_request.method == 'GET' and
           http_request.headers.get('upgrade', '').lower() == 'websocket'):

            return await self.handle_websocket_request(http_request)

        return Response(
            status=405,
            text='405: Method Not Allowed',
        )



from lona.types import Symbol


class SORT_ORDER(Symbol):
    FRAMEWORK = Symbol('FRAMEWORK', 100)
    LIBRARY = Symbol('LIBRARY', 200)
    APPLICATION = Symbol('APPLICATION', 300)


class StaticFile:
    def __init__(self, name, path, url=None, sort_order=None, link=True,
                 enabled_by_default=True):

        self.name = name
        self.path = path
        self.url = url
        self.sort_order = sort_order or SORT_ORDER.APPLICATION
        self.link = link
        self.enabled_by_default = enabled_by_default
        self.enabled = True

        # this value is only for the object representation
        # and is meant for debugging
        self.static_url_prefix = ''

    def __repr__(self):
        return '<{}({}, sort_order={}, enabled={}, {}{} -> {})>'.format(
            self.__class__.__name__,
            self.name,
            repr(self.sort_order),
            repr(self.enabled),
            self.static_url_prefix,
            self.url,
            self.path,
        )


class StyleSheet(StaticFile):
    pass


class Script(StaticFile):
    pass



from threading import RLock

from lona.html.abstract_node import AbstractNode
from lona.html.widget import Widget
from lona.protocol import DATA_TYPE
from lona.html.node import Node


class Document:
    def __init__(self):
        self._lock = RLock()
        self.html = None

    @property
    def lock(self):
        return self._lock

    # html ####################################################################
    def get_node(self, node_id, widget_id=None):
        if isinstance(node_id, Node):
            node_id = node_id.id

        value = [None, [], ]
        widget = [None]
        widget_path = []

        def iter_nodes(node):
            if isinstance(node, Widget):
                widget_path.append(node)

                if widget_id and node.id == widget_id:
                    widget[0] = node

            if isinstance(node, (Node, Widget)):
                if node.id == node_id:
                    value[0] = node
                    value[1].extend(widget_path)

                    return

                if node.nodes:
                    for i in node.nodes:
                        iter_nodes(i)

            if isinstance(node, Widget):
                widget_path.pop()

        with self.lock:
            iter_nodes(self.html)

        if widget[0]:
            value[1].append(widget[0])

        return tuple(value)

    def _has_patches(self):
        def has_patches(node):
            if node._has_patches():
                return True

            if hasattr(node, 'nodes'):
                for sub_node in node.nodes:
                    if has_patches(sub_node):
                        return True

            return False

        return has_patches(self.html)

    def _collect_patches(self):
        patches = []
        patched_widgets = []
        widget_path = []

        # find patches and patched widgets
        def add_patches(node):
            node_is_widget = isinstance(node, Widget)

            if node_is_widget:
                widget_path.append(node.id)

            # patched widgets
            if node_is_widget:
                if node.nodes._has_patches():
                    patched_widgets.extend(widget_path)

            elif node._has_patches():
                patched_widgets.extend(widget_path)

            # patches
            if node._has_patches():
                patches.extend(node._get_patches())

            if hasattr(node, 'nodes'):
                for sub_node in node.nodes:
                    add_patches(sub_node)

            if node_is_widget:
                widget_path.remove(node.id)

        add_patches(self.html)

        # sort patches by timestamp
        patches = sorted(patches, key=lambda x: x[0])

        # remove timestamps
        cleaned_patches = []

        for patch in patches:
            cleaned_patches.append(patch[1:])

        # clean list of patched widgets
        cleaned_patched_widgets = []

        for widget_id in patched_widgets[::-1]:
            if widget_id not in cleaned_patched_widgets:
                cleaned_patched_widgets.append(widget_id)

        return [cleaned_patches, cleaned_patched_widgets]

    def serialize(self):
        if not self.html:
            return self.apply('')

        return DATA_TYPE.HTML_TREE, self.html._serialize()

    def apply(self, html):
        if isinstance(html, str) and html is self.html:
            return

        # HTML update
        elif html is self.html:
            if not self._has_patches():
                return

            patches = self._collect_patches()
            self.html._clear_patches()

            return DATA_TYPE.HTML_UPDATE, patches

        # HTML
        else:
            if isinstance(self.html, AbstractNode):
                self.html._set_document(None)

            # node tree
            if isinstance(html, AbstractNode):
                self.html = html

                self.html._set_document(self)
                self.html._clear_patches()

                return self.serialize()

            # HTML string
            self.html = str(html)

            return DATA_TYPE.HTML, html
